/* -*-c-*-
 *
 * Where the fun begins
 *
 * (c) 2004 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of the Python interface to Catacomb.
 *
 * Catacomb/Python is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Catacomb/Python is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Catacomb/Python; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "catacomb-python.h"

/*----- Main code ---------------------------------------------------------*/

static const struct nameval consts[] = {
#define C(x) { #x, x }
  C(FTY_PRIME), C(FTY_BINARY),
  C(PGEN_PASS), C(PGEN_FAIL), C(PGEN_BEGIN), C(PGEN_TRY), C(PGEN_DONE),
  C(PGEN_ABORT),
  C(MPW_MAX),
  C(RAND_IBITS),
  C(PMODE_READ), C(PMODE_VERIFY),
  C(KOPEN_READ), C(KOPEN_WRITE), C(KOPEN_NOFILE),
  C(KEXP_FOREVER), C(KEXP_EXPIRE),
  C(KF_ENCMASK), C(KENC_BINARY), C(KENC_MP), C(KENC_STRUCT),
    C(KENC_ENCRYPT), C(KENC_STRING), C(KENC_EC),
  C(KF_CATMASK), C(KCAT_SYMM), C(KCAT_PRIV), C(KCAT_PUB), C(KCAT_SHARE),
  C(KF_NONSECRET),
  C(KF_BURN), C(KF_OPT),
  C(EC_XONLY), C(EC_YBIT), C(EC_LSB), C(EC_CMPR), C(EC_EXPLY), C(EC_SORT),
  C(X25519_KEYSZ), C(X25519_PUBSZ), C(X25519_OUTSZ),
  C(X448_KEYSZ), C(X448_PUBSZ), C(X448_OUTSZ),
  C(ED25519_KEYSZ), C(ED25519_PUBSZ), C(ED25519_SIGSZ),
    C(ED25519_MAXPERSOSZ),
  C(ED448_KEYSZ), C(ED448_PUBSZ), C(ED448_SIGSZ), C(ED448_MAXPERSOSZ),
#define ENTRY(tag, val, str) C(KERR_##tag),
  KEY_ERRORS(ENTRY)
#undef ENTRY
#undef C
  { 0 }
};

PyObject *mexp_common(PyObject *me, PyObject *arg,
		      size_t efsz,
		      PyObject *(*id)(PyObject *),
		      int (*fill)(void *, PyObject *,
				  PyObject *, PyObject *),
		      PyObject *(*exp)(PyObject *, void *, int),
		      void (*drop)(void *))
{
  int i = 0, j, n, flat;
  PyObject *qq, *x, *y, *z = 0;
  char *v = 0, *vv;

  if (PyTuple_Size(arg) == 1)
    arg = PyTuple_GetItem(arg, 0);
  Py_INCREF(arg);
  if (!PySequence_Check(arg)) TYERR("not a sequence");
  n = PySequence_Size(arg); if (!n) { z = id(me); goto end; }
  x = PySequence_GetItem(arg, 0);
  if (PySequence_Check(x))
    flat = 0;
  else {
    if (n % 2) VALERR("must have even number of arguments");
    n /= 2;
    flat = 1;
  }
  Py_DECREF(x);

  v = xmalloc(n * efsz);
  for (i = j = 0, vv = v; i < n; i++, vv += efsz) {
    if (flat) {
      x = PySequence_GetItem(arg, j++);
      y = PySequence_GetItem(arg, j++);
    } else {
      qq = PySequence_GetItem(arg, j++);
      if (!qq) goto end;
      if (!PySequence_Check(qq) || PySequence_Size(qq) != 2) {
	Py_DECREF(qq);
	TYERR("want a sequence of pairs");
      }
      x = PySequence_GetItem(qq, 0);
      y = PySequence_GetItem(qq, 1);
      Py_DECREF(qq);
    }
    if (!x || !y) goto end;
    if (fill(vv, me, x, y)) {
      Py_DECREF(x);
      Py_DECREF(y);
      if (!PyErr_Occurred())
	PyErr_SetString(PyExc_TypeError, "type mismatch");
      goto end;
    }
    Py_DECREF(x);
    Py_DECREF(y);
  }
  z = exp(me, v, n);

end:
  if (v) {
    for (j = 0, vv = v; j < i; j++, vv += efsz)
      drop(vv);
    xfree(v);
  }
  Py_DECREF(arg);
  return (z);
}

static PyObject *smallprimes(void)
{
  PyObject *v = PyList_New(NPRIME);
  int i;

  for (i = 0; i < NPRIME; i++)
    PyList_SetItem(v, i, PyInt_FromLong(primetab[i]));
  return (v);
}

static PyObject *meth__ego(PyObject *me, PyObject *arg)
{
  char *argv0;
  if (!PyArg_ParseTuple(arg, "s:_ego", &argv0))
    return (0);
  if (strcmp(QUIS, "<UNNAMED>") == 0)
    ego(argv0);
  RETURN_NONE;
}

static PyMethodDef methods[] = {
#define METHNAME(func) meth_##func
  METH	(_ego,			"_ego(ARGV0)")
#undef METHNAME
  { 0 }
};

static void init_random(void)
{
#if PY_MAJOR_VERSION >= 3 || (PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION >= 6)
  char *seed;
  uint32 r;

  if (!Py_HashRandomizationFlag) return;
  seed = getenv("PYTHONHASHSEED");
  if (!seed || strcmp(seed, "random") == 0) r = GR_WORD(&rand_global);
  else r = strtoul(seed, 0, 0);
  if (!r) r = 0xe011f220; /* zero doesn't work well */
  unihash_setkey(&unihash_global, r);
#endif
}

void init_base(void)
{
  PyObject *mod;
  addmethods(methods);
  INIT_MODULES;
  init_random();
  mod = Py_InitModule("catacomb._base", donemethods());
  INSERT_MODULES;
  INSERT("smallprimes", smallprimes());
  setconstants(mod, consts);
}

/*----- That's all, folks -------------------------------------------------*/
